# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#
# Autopilot Functional Test Tool
# Copyright (C) 2015 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Functions for dealing with proxy classes and objects."""


from autopilot.introspection import (
    constants,
    dbus as ap_dbus
)
from autopilot.introspection._xpathselect import get_classname_from_path
from autopilot.utilities import get_debug_logger


def _get_introspection_xml_from_backend(
        backend, reply_handler=None, error_handler=None):
    """Get DBus Introspection xml from a backend.

    :param backend: The backend object to query.
    :param reply_handler: If set, makes a dbus async call, and the result will
        be sent to reply_handler. This must be a callable object.
    :param error_handler: If set, this callable will recieve any errors, and
        the call will be made asyncronously.
    :returns: A string containing introspection xml, if called synchronously.
    :raises ValueError: if one, but not both of 'reply_handler' and
        'error_handler' are set.

    """
    if callable(reply_handler) and callable(error_handler):
        backend.dbus_introspection_iface.Introspect(
            reply_handler=reply_handler,
            error_handler=error_handler,
        )
    elif reply_handler or error_handler:
        raise ValueError(
            "Both 'reply_handler' and 'error_handler' must be set."
        )
    else:
        return backend.dbus_introspection_iface.Introspect()


def _get_proxy_bases_from_introspection_xml(introspection_xml):
    """Return  tuple of the base classes to use when creating a proxy object.

    Currently this works by looking for certain interface names in the XML. In
    the future we may want to parse the XML and perform more rigerous checks.

    :param introspection_xml: An xml string that describes the exported object
        on the dbus backend. This determines which capabilities are present in
        the backend, and therefore which base classes should be used to create
        the proxy object.

    :raises RuntimeError: if the autopilot interface cannot be found.

    """

    bases = []

    if constants.AP_INTROSPECTION_IFACE not in introspection_xml:
        raise RuntimeError("Could not find Autopilot interface.")

    if constants.QT_AUTOPILOT_IFACE in introspection_xml:
        from autopilot.introspection.qt import QtObjectProxyMixin
        bases.append(QtObjectProxyMixin)

    return tuple(bases)


def _get_proxy_object_class_name_and_state(
        backend, reply_handler=None, error_handler=None):
    """Get details about this autopilot backend via a dbus GetState call.

    :param reply_handler: A callable that must accept three positional
        arguments, which correspond to the return value of this function when
        called synchronously.

    :param error_handler: A callable which will recieve any dbus errors, should
        they occur.

    :raises ValueError: if one, but not both of reply_handler and error_handler
        are set.

    :returns: A tuple containing the class name of the root of the
        introspection tree, the full path to the root of the introspection
        tree, and the state dictionary of the root node in the introspection
        tree.

    """
    if callable(reply_handler) and callable(error_handler):
        # Async call:
        # Since we get an array of state, and we only care about the first one
        # we use a lambda to unpack it and get the details we want.
        backend.introspection_iface.GetState(
            "/",
            reply_handler=lambda r: reply_handler(
                *_get_details_from_state_data(r[0])
            ),
            error_handler=error_handler,
        )
    elif reply_handler or error_handler:
        raise ValueError(
            "Both 'reply_handler' and 'error_handler' must be set."
        )
    else:
        # Sync call
        state = backend.introspection_iface.GetState("/")[0]
        return _get_details_from_state_data(state)


def _get_details_from_state_data(state_data):
    """Get details from a state data array.

    Returns class name, path, and state dictionary.
    """
    object_path, object_state = state_data
    return (
        get_classname_from_path(object_path),
        object_path.encode('utf-8'),
        object_state,
    )


def _get_default_proxy_class(name, bases):
    """Return a default custom proxy object class.

    :param name: name of new class
    :param bases: Tuple of base classes to add to the base.
    :returns: custom proxy object class

    """
    get_debug_logger().warning(
        "Generating introspection instance for type '%s' based on generic "
        "class.", name)
    if isinstance(name, bytes):
        name = name.decode('utf-8')
    return type(
        name,
        (ap_dbus.DBusIntrospectionObject,) + bases,
        dict(__generated=True)
    )


def get_proxy_object_class(cpo_or_string, path, dbus_backend):
    """Return the proxy class to use with the correct bases added

    If passed a CPO returns that, if passed a string contructs a default class
    to use instead.

    """

    extension_classes = _get_extension_classes(dbus_backend)

    if _is_cpo_class(cpo_or_string):
        _apply_mixin_bases(cpo_or_string, extension_classes)
        return cpo_or_string
    else:
        cpo_name = _get_applicable_class_name(cpo_or_string, path)
        return _get_default_proxy_class(cpo_name, extension_classes)


def _get_extension_classes(dbus_backend):
    """Return any extension classes that this object will need."""

    intro_xml = _get_introspection_xml_from_backend(dbus_backend)
    try:
        return _get_proxy_bases_from_introspection_xml(intro_xml)
    except RuntimeError as e:
        e.args = (
            "Could not find Autopilot interface on dbus address '%s'."
            % dbus_backend,
        )


def _is_cpo_class(cpo_or_string):
    return cpo_or_string is not None and not isinstance(cpo_or_string, str)


def _get_applicable_class_name(cpo_name, path):
    """Return correct name for class.

    It is possible that the cpo_name is '*' which is what was used in the
    query. Determine what the actual name should be.

    """
    if cpo_name is None or cpo_name == '*':
        cls_name = get_classname_from_path(path)
        return cls_name.decode('utf-8')
    else:
        return cpo_name


def _apply_mixin_bases(target_class, mixin_classes):
    to_add = tuple(set(mixin_classes).difference(target_class.__bases__))
    target_class.__bases__ += to_add
